/* ppm handling functions */

#include <stdio.h>
#include "mat_util.h"
#include "ppm.h"
#include <string.h>

int load_ppm_file(char * file_name, Var * image)
{
  int magic_number;
  int c;
  unsigned int garbage;
  unsigned char newchar;
  int width, height, depth, channels;
  int num_got;

  unsigned int pixelr;
  FILE * file_ptr;
  unsigned int row_index;
  char *string1, *string2;

  file_ptr = fopen(file_name, "r");
  if(file_ptr == NULL)
    {
      printf("\n  ERROR: file '%s' not found\n", file_name );
      return -1;
    }

  /* - - GET INFO FROM PPM HEADER - - */
  fscanf(file_ptr,"P%x", &magic_number);

  /* - kill a comment - */
  c = getc(file_ptr);
  while (1)
    {
      if (c == '#')
	{
	  while (1)
	    {
	      c = getc(file_ptr);
	      if (c == '\n' || c == EOF)
		break;    
	    }
	}
      if (c==EOF)
	return 0;
      if (c>='0' && c<='9')
	{
	  /* rewind file pointer 1 char position */
	  fseek(file_ptr, -1, SEEK_CUR); 
	  break;  
	}
      /* see if we are getting garbage (non-whitespace) */
      if (c!=' ' && c!='\t' && c!='\r' && c!='\n' && c!=',')
	garbage=1;
      
      c = getc(file_ptr);
    }

  fscanf(file_ptr,"%d %d %d", &width, &height, &depth);

  /* Read off the two extra fields in plms and save for later. */
  if (magic_number == 9 || magic_number == 10) 
      fscanf(file_ptr,"%d %f", &image->plm_field_1, &image->plm_field_2);

/*  printf("\n  width:%d, height:%d, depth:%d, MN:%d\n\n", width,
            height, depth, magic_number);*/

  /* - - set Variable properties M,N, channels, var_type and name - - */
  image->M = height;
  image->N = width;
  image->max_value = depth;
  image->magic_number = magic_number;

  if(magic_number == 2 || magic_number == 5 || magic_number == 7 || magic_number == 9)
    image->channels = GREY_SCALE;
  else if(magic_number == 3 || magic_number == 6 || magic_number == 8 || magic_number == 10)
    image->channels = RGB24;

  channels = image->channels;
  image->var_type = IMAGE;

  /* set matrix name from input file name */
  /* . in file name incorrectly sets the name */
  string1 = (char *)calloc(200, sizeof(char));
  string2 = (char *)calloc(200, sizeof(char));
  
/*  fflush(stdout);*/

  string1 = strrchr(file_name, '/');
  if(string1 == NULL)
    {
      /* cut off .pgm and set name */
      strcpy(string2, file_name);
      sscanf(string2, "%[^.]s", string2);
      image->name = string2;
    }
  else
    {
      sscanf(string1, "/%s.pgm", string2);
      sscanf(string2, "%[^.]s", string2);
/*      fflush(stdout);*/

      image->name = string2;
    }
  /* - - freeing string1 caused an error REMOVED 11/2/94 - - */

  /* - - - alloc memory for picture - - - */
  image->image = (Image *) calloc(width*height*channels, sizeof(Image));

  fread(&pixelr, 1, 1, file_ptr);   /* cut off 10 that keeps showing up */

  /* Read in raw data. */
  if(magic_number == 2 || magic_number == 3 || magic_number == 5 || magic_number == 6)
    {
      for(row_index = 0; row_index < height*width*channels; row_index++)
	{
	  fread(&newchar, sizeof(unsigned char), 1, file_ptr);
	  image->image[row_index] = (Image)newchar;
	}
      num_got = row_index;
    }
  else if(magic_number == 7 || magic_number == 8 || magic_number == 9 || magic_number == 10)
    {	   
      num_got = fread(image->image, sizeof(Image), width*height*channels, file_ptr);
    }
  else
    { 
      printf("load_ppm_file: Image %s is of an unknown file type\n", file_name);
      fclose(file_ptr);
      return -1; 
    }

  printf("num got: %d\n", num_got);
  fclose(file_ptr);

  return 0;
}


/* ====================================================================== */
int save_pgm_file(Var * image, char * filename)
{
  FILE * file_ptr;
  long num;
  int magic_number = image->magic_number;
  int i;

  file_ptr = fopen(filename, "w");
  if(file_ptr == NULL)
    {
      printf("\n  ERROR: unable to create file '%s' \n", filename );
      return(-1);
    }

  /* - - Create file header - - */
  if (magic_number == 10)
    fprintf(file_ptr, "PA\n");
  else
    fprintf(file_ptr, "P%d\n", magic_number);
  fprintf(file_ptr, "# created by ___________ \n");
  fprintf(file_ptr, "%d %d\n", image->N, image->M);
  fprintf(file_ptr, "255\n");

  if(magic_number == 9 || magic_number == 10)
  {
    fprintf(file_ptr, "%d\n", image->plm_field_1);
    fprintf(file_ptr, "%f\n", image->plm_field_2);
  }

  /* - - write raw data - - - */
  if(magic_number == 7 || magic_number == 8 || magic_number == 9 || magic_number == 10) {
    num = fwrite(image->image, sizeof(Image), image->M*image->N*image->channels, file_ptr);
    if (num != (size_t)(image->M*image->N*image->channels)) {
      fclose(file_ptr);
      fprintf(stderr, "Not enough elements output: %ld\n", num);
    }
  }
  else if(magic_number == 2 || magic_number == 3 || magic_number == 5 || magic_number == 6)
    {
      for(i = 0; i < image->M * image->N * image->channels; i++)
	if (image->image[i] >= NaNChar) 
	  fprintf(file_ptr, "%c", (unsigned char)NaNChar);
	else
	  fprintf(file_ptr, "%c", (unsigned char)image->image[i]);
    }
  
  fclose(file_ptr);
  return(1);
}


/* ============================================================ */
/* ============================================================ */
/* ============================================================ */
int get_image_dimensions(char * filename, int * width, int * height)
{
  int magic_number;
  int c;                        /* for reading comments */
  unsigned int garbage;

  FILE * file_ptr;

  file_ptr = fopen(filename, "r");
  if(file_ptr == NULL)
    {
      printf("\n  ERROR: Could not open file '%s' to find dimensions\n", filename );
      return(-1);
    }

  /* - - GET INFO FROM PPM HEADER - - */
  fscanf(file_ptr,"P%d", &magic_number);

  /* - kill a comment - */
  c = getc(file_ptr);
  while (1)
    {
      if (c == '#')
	{
	  while (1)
	    {
	      c = getc(file_ptr);
	      if (c == '\n' || c == EOF)
		break;    
	    }
	}
      if (c==EOF)
	return 0;
      if (c>='0' && c<='9')
	{
	  /* rewind file pointer 1 char position */
	  fseek(file_ptr, -1, SEEK_CUR); 
	  break;  
	}
      /* see if we are getting garbage (non-whitespace) */
      if (c!=' ' && c!='\t' && c!='\r' && c!='\n' && c!=',')
	garbage=1;
      
      c = getc(file_ptr);
    }
  
  fscanf(file_ptr,"%d %d", width, height);
  return(0);
}

