#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "libgimp/gimp.h"
#include "gtk/gtk.h"

//Lightspace test
//howto compile:   gcc -lgtk -lgdk -lgimp test.c -o test

static void query(void);
static void run(char *name, int nparams, GParam *param, int *nreturn_vals, GParam **return_vals);

//Helper Funcs...
float normalizeByte(unsigned char x);
unsigned char byteizeNormalFloat(float x);

GPlugInInfo PLUG_IN_INFO =
{
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};

MAIN ()

static void query ()
{
  static GParamDef args[] =
  {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE, "image", "Input image" },
    { PARAM_DRAWABLE, "drawable", "Input drawable" },
    { PARAM_IMAGE, "image", "1st Image"},
    { PARAM_IMAGE, "image", "2nd Image"}
  };
  static GParamDef *return_vals = NULL;
  static int nargs = sizeof (args) / sizeof (args[0]);
  static int nreturn_vals = 0;

  gimp_install_procedure ("plug_in_LightspaceProto",
			  "Lightspace Prototype",
			  "This plugin does some neat image manipulation",
			  "Maneesh Yadav",
			  "Maneesh Yadav",
			  "2000",
			  "<Image>/lightspace",
			  "RGB*, GRAY*, INDEXED*",
			  PROC_PLUG_IN,
			  nargs, nreturn_vals,
			  args, return_vals);
}



static void run (char *name, int nparams, GParam  *param, int *nreturn_vals, GParam **return_vals)
{
 static GParam values[1];  
 gint32 imageid,drawableid,newlayerid,maskid;
 GPixelRgn prl1,prl2,prN;
 gpointer pr;
 GDrawable *drawablel1,*drawablel2,*drawableN;
 GStatusType status = STATUS_SUCCESS; // For now..add status later..
 int width,height;
 int i,j;
 float pixel1[4];
 float pixel2[4];
 float pixel3[4];
 gint32 l1,l2;
 
 
 int nlayers;
 gint* layerlist;
 
 *nreturn_vals = 0;
 *return_vals = values;
 values[0].type = PARAM_STATUS;
 values[0].data.d_status = STATUS_SUCCESS;
 imageid=param[1].data.d_image;
 
 width = gimp_image_width(imageid);
 height= gimp_image_height(imageid);
 
 layerlist=gimp_image_get_layers(imageid,&nlayers);

 l1=layerlist[0];
 l2=layerlist[1];

 newlayerid=gimp_layer_new(imageid,"results",width,height,RGBA_IMAGE,100.0,NORMAL_MODE);
 gimp_image_add_layer(imageid,newlayerid,2);


 drawablel1=gimp_drawable_get(l1);
 drawablel2=gimp_drawable_get(l2);
 drawableN=gimp_drawable_get(newlayerid);
 
 gimp_pixel_rgn_init(&prl1, drawablel1,0,0,width,height,TRUE,FALSE);
 gimp_pixel_rgn_init(&prl2, drawablel2,0,0,width,height,TRUE,FALSE);
 gimp_pixel_rgn_init(&prN, drawableN,0,0,width,height,TRUE,TRUE);



 for (pr = gimp_pixel_rgns_register(3, &prl1, &prl2, &prN); pr != NULL;  pr = gimp_pixel_rgns_process(pr))
 {
  for(i=0; i<(prN.w*prN.h*prN.bpp); i+=4)
  {
   

   pixel1[0]=normalizeByte(prl1.data[i]);
   pixel1[1]=normalizeByte(prl1.data[i+1]);
   pixel1[2]=normalizeByte(prl1.data[i+2]);
   pixel1[3]=normalizeByte(prl1.data[i+3]);
  
   pixel2[0]=normalizeByte(prl2.data[i]);
   pixel2[1]=normalizeByte(prl2.data[i+1]);
   pixel2[2]=normalizeByte(prl2.data[i+2]);
   pixel2[3]=normalizeByte(prl2.data[i+3]); 
   
   pixel3[0]=sqrt((pixel1[0]*pixel1[0]+pixel2[0]*pixel2[0]));
   pixel3[1]=sqrt((pixel1[1]*pixel1[1]+pixel2[1]*pixel2[1]));
   pixel3[2]=sqrt((pixel1[2]*pixel1[2]+pixel2[2]*pixel2[2]));
   pixel3[3]=sqrt((pixel1[3]*pixel1[3]+pixel2[3]*pixel2[3]));
   
   for(j=0;j<4;j++)
    if(pixel3[j]>1) pixel3[j]=1;

   prN.data[i]  =byteizeNormalFloat(pixel3[0]);
   prN.data[i+1]=byteizeNormalFloat(pixel3[1]);
   prN.data[i+2]=byteizeNormalFloat(pixel3[2]);
   prN.data[i+3]=byteizeNormalFloat(pixel3[3]);
  }
 }
 
 gimp_drawable_flush (drawablel1); 
 gimp_drawable_flush (drawablel2);          
 gimp_drawable_flush (drawableN);
 gimp_drawable_merge_shadow (drawableN->id, TRUE);
 gimp_drawable_update (drawablel1->id, 0, 0, width, height);
 gimp_drawable_update (drawablel2->id, 0, 0, width, height);
 gimp_drawable_update (drawableN->id, 0, 0, width, height); 
 gimp_displays_flush();
 gimp_drawable_detach(drawablel1);
 gimp_drawable_detach(drawablel2); 
 gimp_drawable_detach(drawableN); 

 return;
}

float normalizeByte(unsigned char x)
{
 return ((float)x/255.0);
}

unsigned char byteizeNormalFloat(float x)
{
 return (unsigned char)((x)*255);
}
