#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "libgimp/gimp.h"
#include "libgimp/gimpmenu.h"
#include "gtk/gtk.h"

//Lightspace test
//howto compile:   gcc -lgtk -lgdk -lgimp test.c -o test

//Helper Funcs...
float normalizeByte(unsigned char x);
unsigned char byteizeNormalFloat(float x);

LightSpaceState* newLightSpaceState();
void deleteLightSpaceState(LightSpaceState* ls);
int hasInputLayerInLightSpaceState(int layerID, LightSpaceState* ls);
void addInputLayerToLightSpaceState(int layerID, LightSpaceState* ls);
LightSpaceDialogState* newLightSpaceDialogState();

gint32 lsDialog(LightSpaceState* ls);

static gint32 layerConstraintCallback(gint32 image_id, gint32 drawable_id, gpointer data);

static void layerSelectMenuCallback(gint32 id, gpointer data);
static void layerInputtMenuCallback(gint32 id, gpointer data);


static void addButtonCallback (GtkWidget *widget, gpointer data);
static void dialogCloseCallback (GtkWidget *widget, gpointer data);

#define PLUG_IN_NAME "LightSpace"
#define PDB_PLUGIN_NAME "plug_in_LightspaceProto"

typedef struct _LightSpaceState
{
 gint32   nLayers;
 gint32	  nInputLayers;
 gint32*  layerList;
 gint32*  inputFlags;
 double** rgbWeights;
 LightSpaceDialogState* dialogState;
} LightSpaceState;


typedef struct _LightSpaceDialogState
 gint32 layerSelect;
 gint32 layerInput;
} LightSpaceDialogState;

 
static void query(void);
static void run(char *name, int nparams, GParam *param, int *nreturn_vals, GParam **return_vals);


GPlugInInfo PLUG_IN_INFO =
{
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};

MAIN ()

static void query ()
{
  static GParamDef args[] =
  {
    { PARAM_INT32, "run_mode", "Interactive, non-interactive" },
    { PARAM_IMAGE, "image", "Input image" },
    { PARAM_DRAWABLE, "drawable", "Input drawable" },
    { PARAM_IMAGE, "image", "1st Image"},
    { PARAM_IMAGE, "image", "2nd Image"}
  };
  static GParamDef *return_vals = NULL;
  static int nargs = sizeof (args) / sizeof (args[0]);
  static int nreturn_vals = 0;

  gimp_install_procedure (PDB_PLUGIN_NAME,
			  PLUG_IN_NAME,
			  "This plugin does some neat image manipulation",
			  "Maneesh Yadav",
			  "Maneesh Yadav",
			  "2000",
			  "<Image>/lightspace",
			  "RGB*, GRAY*, INDEXED*",
			  PROC_PLUG_IN,
			  nargs, nreturn_vals,
			  args, return_vals);
}



static void run (char *name, int nparams, GParam  *param, int *nreturn_vals, GParam **return_vals)
{
 static GParam values[1];  
 LightSpaceParams parameters;
 gint32 imageid,drawableid,newlayerid,maskid;
 GPixelRgn prl1,prl2,prN;
 gpointer pr;
 GDrawable *drawablel1,*drawablel2,*drawableN;
 GStatusType status = STATUS_SUCCESS; // For now..add status later..
 int width,height;
 int i,j;
 float pixel1[4];
 float pixel2[4];
 float pixel3[4];
 gint32 l1,l2;
 LightSpaceState* ls;
 
 gint nlayers;
 gint* layerlist;
 
 
 GPixelRgn* pixelReigons;
 GDrawable* drawables;
  

 ls=newLightSpaceState();

 *nreturn_vals = 0;
 *return_vals = values;
 values[0].type = PARAM_STATUS;
 values[0].data.d_status = STATUS_SUCCESS;
 
  switch (param[0].data.d_int32) {
  case RUN_INTERACTIVE:
    /*  Possibly retrieve data  */
//    gimp_get_data(PDB_PLUGIN_NAME, &parameters);

    /*  Get information from the dialog  */
    if (!lsDialog (ls))
      return;

    break;

  case RUN_NONINTERACTIVE:
    break;

  case RUN_WITH_LAST_VALS:
    break;
    
  default:
    break;
  }
 
 
 imageid=param[1].data.d_image;
 
 width = gimp_image_width(imageid);
 height= gimp_image_height(imageid);
 
 layerlist=gimp_image_get_layers(imageid,&nlayers);

 l1=layerlist[0];
 l2=layerlist[1];
 
 newlayerid=gimp_layer_new(imageid,"results",width,height,RGBA_IMAGE,100.0,NORMAL_MODE);
 gimp_image_add_layer(imageid,newlayerid,2);


 for(i=0;i<ls->nLayers; i++)
 {
  i
 drawablel1=gimp_drawable_get(l1);
 drawablel2=gimp_drawable_get(l2);
 drawableN=gimp_drawable_get(newlayerid);
 
 gimp_pixel_rgn_init(&prl1, drawablel1,0,0,width,height,TRUE,FALSE);
 gimp_pixel_rgn_init(&prl2, drawablel2,0,0,width,height,TRUE,FALSE);
 gimp_pixel_rgn_init(&prN, drawableN,0,0,width,height,TRUE,TRUE);


 for(i=0; i<ls->nInputLayers; i++)
  tileIterators[i] = gimp_pixel_rgns_register(1,pr[i]);
 
 while(tileIterators[0] != NULL)
 {
  
  for(i=0; i<ls->nInputLayers; i++)
   tileIterators[i]=gimp_pixel_rgns_process(tileIterators[i]);
 }
 for (pr = gimp_pixel_rgns_register(3, &prl1, &prl2, &prN); pr != NULL;  pr = gimp_pixel_rgns_process(pr))
 {
  for(i=0; i<(prN.w*prN.h*prN.bpp); i+=4)
  {
   

   pixel1[0]=normalizeByte(prl1.data[i]);
   pixel1[1]=normalizeByte(prl1.data[i+1]);
   pixel1[2]=normalizeByte(prl1.data[i+2]);
   pixel1[3]=normalizeByte(prl1.data[i+3]);
  
   pixel2[0]=normalizeByte(prl2.data[i]);
   pixel2[1]=normalizeByte(prl2.data[i+1]);
   pixel2[2]=normalizeByte(prl2.data[i+2]);
   pixel2[3]=normalizeByte(prl2.data[i+3]); 
   
   pixel3[0]=sqrt(1.0f/(pixel1[0]*pixel1[0]+pixel2[0]*pixel2[0])/2);
   pixel3[1]=sqrt(1.0f/(pixel1[1]*pixel1[1]+pixel2[1]*pixel2[1])/2);
   pixel3[2]=sqrt(1.0f/(pixel1[2]*pixel1[2]+pixel2[2]*pixel2[2])/2);
   pixel3[3]=sqrt(1.0f/(pixel1[3]*pixel1[3]+pixel2[3]*pixel2[3])/2);
   
   prN.data[i]  =byteizeNormalFloat(pixel3[0]);
   prN.data[i+1]=byteizeNormalFloat(pixel3[1]);
   prN.data[i+2]=byteizeNormalFloat(pixel3[2]);
   prN.data[i+3]=byteizeNormalFloat(pixel3[3]);
  }
 }
 
 gimp_drawable_flush (drawablel1); 
 gimp_drawable_flush (drawablel2);          
 gimp_drawable_flush (drawableN);
 gimp_drawable_merge_shadow (drawableN->id, TRUE);
 gimp_drawable_update (drawablel1->id, 0, 0, width, height);
 gimp_drawable_update (drawablel2->id, 0, 0, width, height);
 gimp_drawable_update (drawableN->id, 0, 0, width, height); 
 gimp_displays_flush();
 gimp_drawable_detach(drawablel1);
 gimp_drawable_detach(drawablel2); 
 gimp_drawable_detach(drawableN); 
 
 deleteLightSpaceState(ls);
 
 return;
}

gint32 lsDialog(LightSpaceState* ls) 
{
  gint      argc;
  gchar     **argv;
 
  GtkWidget *dialog,*topTable,*sourceTable, *layerInputOptionMenu, *i,*oLayerMenu;
  GtkWidget *addButton, *removeButton;
  
  guchar    *color_cube;

  argc = 1;
  argv = g_new(gchar *, 1);
  argv[0] = g_strdup(PLUG_IN_NAME);

  gtk_init(&argc, &argv);
  gtk_rc_parse(gimp_gtkrc());

  gdk_set_use_xshm(gimp_use_xshm());

  gtk_preview_set_gamma(gimp_gamma());
  gtk_preview_set_install_cmap(gimp_install_cmap());
  color_cube = gimp_color_cube();
  gtk_preview_set_color_cube(color_cube[0],
			     color_cube[1],
			     color_cube[2],
			     color_cube[3]);

  gtk_widget_set_default_visual(gtk_preview_get_visual());
  gtk_widget_set_default_colormap(gtk_preview_get_cmap());

  dialog = gtk_dialog_new();
  gtk_widget_show(dialog);

  gtk_window_set_title(GTK_WINDOW(dialog), PLUG_IN_NAME);
  gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
  gtk_container_border_width(GTK_CONTAINER(dialog), 0);
  gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
		     (GtkSignalFunc)dialogCloseCallback,NULL);
  /* topTable */
  topTable = gtk_table_new(2, 2, FALSE);
  gtk_container_border_width(GTK_CONTAINER(topTable), 6);
  gtk_table_set_row_spacings(GTK_TABLE(topTable), 4);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), topTable,
		     FALSE, FALSE, 0);
  gtk_widget_show(topTable);



//Source Picking
  sourceTable = gtk_table_new(1, 2, FALSE);
  gtk_container_border_width(GTK_CONTAINER(sourceTable), 2);
  gtk_table_attach(GTK_TABLE(topTable), sourceTable, 0, 1, 1, 2,
		   GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
  gtk_widget_show(sourceTable);

  layerSelectOptionMenu=gtk_option_menu_new();
  layerSelectMenu=gimp_layer_menu_new(
                                  (GimpConstraintFunc)layerSelectConstraintCallback,
				  (GimpMenuCallback)layerSelectMenuCallback,
				  (gpointer)ls,
				  0
				);

  
  gtk_table_attach(GTK_TABLE(sourceTable),optionMenu, 0, 1, 0, 1,
                   GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0); 

  gtk_option_menu_set_menu(GTK_OPTION_MENU(optionMenu),layerSelectMenu);
  gtk_widget_show(optionMenu);
  
  addButton = gtk_button_new_with_label("Add Layer");
  GTK_WIDGET_SET_FLAGS(addButton, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(addButton), "clicked",
		     (GtkSignalFunc)addButtonCallback,
		     ls);
		     
  gtk_table_attach(GTK_TABLE(sourceTable), addButton, 1, 2, 0, 1,
                   GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  
  gtk_widget_show(addButton);

  
//Layer List
  inputLayerTable = gtk_table_new(1, 2, FALSE);
  gtk_container_border_width(GTK_CONTAINER(sourceTable), 2);
  gtk_table_attach(GTK_TABLE(topTable), inputLayerTable, 1, 2, 1, 2,
		   GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 0, 0);
  gtk_widget_show(inputLayerTable);

  layerInputOptionMenu=gtk_option_menu_new();
  layerInputMenu=gimp_layer_menu_new(
                                  (GimpConstraintFunc)layerConstraintCallback,
				  (GimpMenuCallback)layerMenuCallback,
				  (gpointer)ls,
				  0
				);

  
  gtk_table_attach(GTK_TABLE(sourceTable),optionMenu, 0, 1, 0, 1,
                   GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0); 

  gtk_option_menu_set_menu(GTK_OPTION_MENU(optionMenu),layerInputMenu);
  gtk_widget_show(optionMenu);
  
  removeButton = gtk_button_new_with_label("Remove Layer");
  GTK_WIDGET_SET_FLAGS(addButton, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(addButton), "clicked",
		     (GtkSignalFunc)removeButtonCallback,
		     ls);
		     
  gtk_table_attach(GTK_TABLE(sourceTable), removeButton, 1, 2, 0, 1,
                   GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
  
  gtk_widget_show(removeButton);

  
  gtk_main();
 
  gdk_flush();

 

  g_free(argv[0]);
  g_free(argv);
  
  return 1;
}


float normalizeByte(unsigned char x)
{
 return ((float)x/255.0);
}

unsigned char byteizeNormalFloat(float x)
{
 return (unsigned char)((x)*255);
}


LightSpaceState* newLightSpaceState()
{
 int i;
  
 ls=g_new(LightSpaceState,1);
 
 ls->nInputLayers=0;
 
 ls->layerList=gimp_image_get_layers(imageid,&(ls->nlayers));
 rgbWeights=g_new(double*,nlayers)
 
 inputFlags=g_new0(gint32, nlayers);
 
 for(i=0; i<nlayers; i++)
 {
  rgbWeights[i]=g_new(double, 3);
  rgbWeights[i][0]=1.0;
  rgbWeights[i][1]=1.0;
  rgbWeights[i][2]=1.0;
 }
 
 ls->dialogState=g_new(LightSpaceDialogState,1);
 initLightSpaceDialogState(ls->dialogState);
}

void deleteLightSpaceState(LightSpaceState* ls)
{
 g_free(ls->layerList);
 g_free(ls->inputFlags);
 
 for(i=0; i<nlayers; i++)
  g_free(ls->rgbWeights[i]);

 g_free(ls->rgbWeights);
 g_free(ls->dialogState);
}

LightSpaceDialogState* newLightSpaceDialogState()
{
 dialogState->layerInput=0;
 dialogState->layerSelect=0;
}

int hasInputLayerInLightSpaceState(int layerID, LightSpaceState* ls)
{
 int i;
 
 for(i=0; i<ls->nLayers; i++)
  if(ls->layerList[i] == layerID)
   if(ls->inputFlags[i]==TRUE)
    return TRUE;
}

void addInputLayerToLightSpaceState(int layerID, LightSpaceState* ls)
{
 int i;
 int index;
 
 for(i=0; i<ls->nLayers; i++)
  if(ls->layerList[i] == layerID)
   index==i;
 
 ls->inputFlags[index]=TRUE;
 ls->nInputLayers++;
}

void removeInputLayerFromLightSpaceState(int layerID, LightSpaceState* ls)
{
 int i;
 int index;

 for(i=0; i<ls->nLayers; i++)
  if(ls->layerList[i] == layerID)
   index==i;

 ls->inputFlags[index]=FALSE;
 ls->nInputLayers--;
}



static gint32 layerAddConstraintCallback(gint32 image_id, gint32 drawable_id, gpointer data)
{
 return TRUE;
}

static gint32 layerRemoveConstraintCallback(gint32 image_id, gint32 drawable_id, gpointer data)
{
// LightSpaceState* ls=(LightSpaceState*) data;
// if(hasInputLayerInLightSpaceState(drawable_id, ls))
  return TRUE;
}

static void layerSelectMenuCallback(gint32 id, gpointer data)
{
 LightSpaceState* ls=(LightSpaceState*)data;
 state->dialogState->layerSelect=id;
 printf("Selected Input Layer: %d\n", state->dialogState->layerSelect);
}

static void layerInputMenuCallback(gint32 id, gpointer data)
{
 LightSpaceState* ls=(LightSpaceState*)data;
 state->dialogState->layerInput=id;
 printf("Edit Input Layer: %d\n", state->dialogState->layerInput);
}

static void addButtonCallback (GtkWidget *widget, gpointer data)
{
  LightSpaceState* ls=(LightSpaceState*)data;
  printf("Added Input Layer: %d\n", state->dialogState->layerSelect);
}

static void removeButtonCallback (GtkWidget *widget, gpointer data)
{
  LightSpaceState* ls=(LightSpaceState*)data;
  printf("Removed Input Layer: %d\n", state->dialogState->layerInput);
}

static void dialogCloseCallback (GtkWidget *widget, gpointer data)
{
  gtk_main_quit ();
}
